/*---------------------------------------------------------------------------*\
 ##   ####  ######     |
 ##  ##     ##         | Copyright: ICE Stroemungsfoschungs GmbH
 ##  ##     ####       |
 ##  ##     ##         | http://www.ice-sf.at
 ##   ####  ######     |
-------------------------------------------------------------------------------
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 1991-2008 OpenCFD Ltd.
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is based on OpenFOAM.

    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

Contributors/Copyright:
    2012-2013 Bernhard F.W. Gschaider <bgschaid@ice-sf.at>

 SWAK Revision: $Id$
\*---------------------------------------------------------------------------*/

#include "gameOfLifeFunction.H"
#include "FieldValueExpressionDriver.H"

#include "addToRunTimeSelectionTable.H"

#include "emptyFvPatchField.H"
#include "symmetryFvPatchField.H"

namespace Foam {

defineTypeNameAndDebug(gameOfLifeFunction,0);
addNamedToRunTimeSelectionTable(FieldValuePluginFunction, gameOfLifeFunction , name, gameOfLife_step);

// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

gameOfLifeFunction::gameOfLifeFunction(
    const FieldValueExpressionDriver &parentDriver,
    const word &name,
    const string &arguments
):
    FieldValuePluginFunction(
        parentDriver,
        name,
        word("volScalarField"),
        arguments
    )
{
    setConwayNumbers();
}

gameOfLifeFunction::gameOfLifeFunction(
    const FieldValueExpressionDriver &parentDriver,
    const word &name
):
    FieldValuePluginFunction(
        parentDriver,
        name,
        word("volScalarField"),
        string("oldState internalField volScalarField")
    )
{
    setConwayNumbers();
}

// * * * * * * * * * * * * * * * * Destructor  * * * * * * * * * * * * * * * //


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

void gameOfLifeFunction::setConwayNumbers()
{
    survivingNumbers_.clear();
    survivingNumbers_.insert(2);
    survivingNumbers_.insert(3);
    birthNumbers_.clear();
    birthNumbers_.insert(3);
}

void gameOfLifeFunction::setArgument(
    label index,
    const string &content,
    const CommonValueExpressionDriver &driver
) {
    if(index==0) {
        oldState_.set(
            new volScalarField(
                dynamic_cast<const FieldValueExpressionDriver &>(
                    driver
                ).getResult<volScalarField>()
            )
        );
    } else {
        FatalErrorIn("gameOfLifeFunction::setArgument("
        "label index,"
        "const string &content,"
        "const CommonValueExpressionDriver &driver)")
            << "This should not happen. Expecting index==0. Got " << index
                << endl
                << exit(FatalError);
    }
}

autoPtr<volScalarField> gameOfLifeFunction::doStep(const volScalarField &old) {
    autoPtr<volScalarField> pField(
        new volScalarField(
            IOobject(
                "golNextStep",
                mesh().time().timeName(),
                mesh(),
                IOobject::NO_READ,
                IOobject::NO_WRITE
            ),
            mesh(),
            dimensionedScalar("dead",dimless,0),
            "zeroGradient"
        )
    );
    volScalarField &field=pField();

    const cellList &cells=old.mesh().cells();
    const labelList &owner=old.mesh().faceOwner();
    const labelList &neighbor=old.mesh().faceNeighbour();

    forAll(old,cellI) {
        bool alive=isAlive(old[cellI]);
        label nOthers=0;
        label nAlive=0;
        const cell &c=cells[cellI];
        forAll(c,i) {
            const label faceI=c[i];
            bool oAlive=false;
            bool isOK=true;
            if(faceI<field.mesh().nInternalFaces()) {
                if(owner[faceI]==cellI) {
                    oAlive=isAlive(old[neighbor[faceI]]);
                } else if(neighbor[faceI]==cellI) {
                    oAlive=isAlive(old[owner[faceI]]);
                } else {
                    FatalErrorIn("gameOfLifeFunction::doStep")
                        << "Cell "
                            << cellI << " is neither owner nor neighbor of face "
                            << faceI << ". Confused"
                            << endl
                            << exit(FatalError);
                }
            } else {
                label patchI=old.mesh().boundaryMesh().whichPatch(faceI);
                if(patchI<0) {
                    FatalErrorIn("gameOfLifeFunction::doStep")
                        << "No patch found for face " << faceI
                            << endl
                            << exit(FatalError);
                }
                const fvPatchField<scalar> &pf=old.boundaryField()[patchI];
                label ind=faceI-pf.patch().patch().start();
                if(pf.patch().coupled()) {
                    // We're looking at the other side
                    oAlive=isAlive(
                        dynamicCast<const coupledFvPatchField<scalar> >(
                            pf
                        ).patchNeighbourField()()[ind]
                    ) ;
                } else if(isA<emptyFvPatchField<scalar> >(pf))  {
                    //                    Dbug << "empty " << alive << endl;
                    isOK=false;
                } else if(isA<basicSymmetryFvPatchField<scalar> >(pf))  {
                    //                    Dbug << "sy " << alive << endl;
                    oAlive=alive;
                } else {
                    // regular patches are "dead"
                    oAlive=false;
                }
            }

            if(isOK) {
                nOthers++;
                if(oAlive) {
                    nAlive++;
                }
            }
        }
        label n=nAlive;
        //        label n=label((nAlive)*(8./nOthers));
        if(
            (alive && survivingNumbers_.found(n))
            ||
            (!alive && birthNumbers_.found(n))
        ) {
            field[cellI]=1;
        } else {
            field[cellI]=0;
        }
    }

    return pField;
}

void gameOfLifeFunction::doEvaluation()
{
    result().setObjectResult(doStep(oldState()));
}

// * * * * * * * * * * * * * * * Friend Operators  * * * * * * * * * * * * * //

} // namespace

// ************************************************************************* //
